/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate model Data.

  Copyright (C) 2003-2020 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <cdi.h>

#include "functs.h"
#include "process_int.h"
#include "percentiles.h"

using funcType1 = double (size_t, const Varray<double> &);
using funcTypeMV1 = double (size_t, const Varray<double> &, double);
using funcType2 = double (size_t, const Varray<double> &, const Varray<double> &, double);
using funcTypeMV2 = double (size_t, const Varray<double> &, const Varray<double> &, double);
using funcType3 = double (size_t, const Varray<double> &, const Varray<double> &, size_t, double);
using funcType4 = double (size_t, const Varray<double> &, size_t, double);

template <typename T>
static void
varrayCopyMer(size_t i, size_t nx, size_t ny, const Varray<T> &v1, Varray<double> &v2)
{
  for (size_t j = 0; j < ny; j++) v2[j] = v1[j * nx + i];
}

static void
fieldMerKernel1(const Field &field1, Field &field2, funcType1 func, funcTypeMV1 funcMV)
{
  size_t rnmiss = 0;
  const auto nmiss = field1.nmiss;
  const auto missval = field1.missval;
  const auto nx = gridInqXsize(field1.grid);
  const auto ny = gridInqYsize(field1.grid);

  Varray<double> v(ny);

  for (size_t i = 0; i < nx; ++i)
    {
      if (field1.memType == MemType::Float)
        varrayCopyMer(i, nx, ny, field1.vec_f, v);
      else
        varrayCopyMer(i, nx, ny, field1.vec_d, v);

      const auto result = nmiss ? funcMV(ny, v, missval) : func(ny, v);
      if (DBL_IS_EQUAL(result, missval)) rnmiss++;
      field2.vec_d[i] = result;
    }

  field2.nmiss = rnmiss;
}

static void
fieldMerKernel2(const Field &field1, Field &field2, funcType2 func, funcTypeMV2 funcMV)
{
  size_t rnmiss = 0;
  const auto nmiss = field1.nmiss;
  const auto missval = field1.missval;
  const auto nx = gridInqXsize(field1.grid);
  const auto ny = gridInqYsize(field1.grid);

  Varray<double> v(ny), w(ny);

  for (size_t i = 0; i < nx; ++i)
    {
      varrayCopyMer(i, nx, ny, field1.weightv, w);
      if (field1.memType == MemType::Float)
        varrayCopyMer(i, nx, ny, field1.vec_f, v);
      else
        varrayCopyMer(i, nx, ny, field1.vec_d, v);

      const auto result = nmiss ? funcMV(ny, v, w, missval) : func(ny, v, w, missval);
      if (DBL_IS_EQUAL(result, missval)) rnmiss++;
      field2.vec_d[i] = result;
    }

  field2.nmiss = rnmiss;
}

static void
fieldMerKernel3(const Field &field1, Field &field2, funcType3 func)
{
  size_t rnmiss = 0;
  const auto nmiss = field1.nmiss;
  const auto missval = field1.missval;
  const auto nx = gridInqXsize(field1.grid);
  const auto ny = gridInqYsize(field1.grid);

  Varray<double> v(ny), w(ny);

  for (size_t i = 0; i < nx; ++i)
    {
      varrayCopyMer(i, nx, ny, field1.weightv, w);
      if (field1.memType == MemType::Float)
        varrayCopyMer(i, nx, ny, field1.vec_f, v);
      else
        varrayCopyMer(i, nx, ny, field1.vec_d, v);

      const auto result = func(ny, v, w, nmiss, missval);
      if (DBL_IS_EQUAL(result, missval)) rnmiss++;
      field2.vec_d[i] = result;
    }

  field2.nmiss = rnmiss;
}

static void
fieldMerKernel4(const Field &field1, Field &field2, funcType4 func)
{
  size_t rnmiss = 0;
  const auto nmiss = field1.nmiss;
  const auto missval = field1.missval;
  const auto nx = gridInqXsize(field1.grid);
  const auto ny = gridInqYsize(field1.grid);

  Varray<double> v(ny);

  for (size_t i = 0; i < nx; ++i)
    {
      if (field1.memType == MemType::Float)
        varrayCopyMer(i, nx, ny, field1.vec_f, v);
      else
        varrayCopyMer(i, nx, ny, field1.vec_d, v);

      const auto result = func(ny, v, nmiss, missval);
      if (DBL_IS_EQUAL(result, missval)) rnmiss++;
      field2.vec_d[i] = result;
    }

  field2.nmiss = rnmiss;
}

static void
field_mer_min(const Field &field1, Field &field2)
{
  fieldMerKernel1(field1, field2, varray_min, varray_min_mv);
}

static void
field_mer_max(const Field &field1, Field &field2)
{
  fieldMerKernel1(field1, field2, varray_max, varray_max_mv);
}

static void
field_mer_range(const Field &field1, Field &field2)
{
  fieldMerKernel1(field1, field2, varray_range, varray_range_mv);
}

static void
field_mer_sum(const Field &field1, Field &field2)
{
  fieldMerKernel1(field1, field2, varray_sum, varray_sum_mv);
}

static void
field_mer_meanw(const Field &field1, Field &field2)
{
  fieldMerKernel2(field1, field2, varray_weighted_mean, varray_weighted_mean_mv);
}

static void
field_mer_avgw(const Field &field1, Field &field2)
{
  fieldMerKernel2(field1, field2, varray_weighted_mean, varray_weighted_avg_mv);
}

static void
field_mer_varw(const Field &field1, Field &field2)
{
  fieldMerKernel3(field1, field2, varray_weighted_var);
}

static void
field_mer_var1w(const Field &field1, Field &field2)
{
  fieldMerKernel3(field1, field2, varray_weighted_var_1);
}

static void
field_mer_stdw(const Field &field1, Field &field2)
{
  size_t rnmiss = 0;
  const auto missval = field1.missval;

  const auto nx = gridInqXsize(field1.grid);

  field_mer_varw(field1, field2);

  for (size_t i = 0; i < nx; i++)
    {
      const auto rstd = varToStd(field2.vec_d[i], missval);
      if (DBL_IS_EQUAL(rstd, missval)) rnmiss++;
      field2.vec_d[i] = rstd;
    }

  field2.nmiss = rnmiss;
}

static void
field_mer_std1w(const Field &field1, Field &field2)
{
  size_t rnmiss = 0;
  const auto missval = field1.missval;

  const auto nx = gridInqXsize(field1.grid);

  field_mer_var1w(field1, field2);

  for (size_t i = 0; i < nx; i++)
    {
      const auto rstd = varToStd(field2.vec_d[i], missval);
      if (DBL_IS_EQUAL(rstd, missval)) rnmiss++;
      field2.vec_d[i] = rstd;
    }

  field2.nmiss = rnmiss;
}

static void
field_mer_kurt(const Field &field1, Field &field2)
{
  fieldMerKernel4(field1, field2, varray_kurt);
}

static void
field_mer_skew(const Field &field1, Field &field2)
{
  fieldMerKernel4(field1, field2, varray_skew);
}

void
field_mer_pctl(const Field &field1, Field &field2, const int p)
{
  size_t rnmiss = 0;
  const auto missval = field1.missval;

  const auto nx = gridInqXsize(field1.grid);
  const auto ny = gridInqYsize(field1.grid);

  Varray<double> v(ny);

  if (field1.nmiss)
    {
      for (size_t i = 0; i < nx; i++)
        {
          size_t k = 0;
          if (field1.memType == MemType::Float)
            {
              for (size_t j = 0; j < ny; j++)
                if (!DBL_IS_EQUAL(field1.vec_d[j * nx + i], missval)) v[k++] = field1.vec_f[j * nx + i];
            }
          else
            {
              for (size_t j = 0; j < ny; j++)
                if (!DBL_IS_EQUAL(field1.vec_d[j * nx + i], missval)) v[k++] = field1.vec_d[j * nx + i];
            }

          if (k > 0)
            {
              field2.vec_d[i] = percentile(v.data(), k, p);
            }
          else
            {
              field2.vec_d[i] = missval;
              rnmiss++;
            }
        }
    }
  else
    {
      for (size_t i = 0; i < nx; i++)
        {
          if (ny > 0)
            {
              if (field1.memType == MemType::Float)
                varrayCopyMer(i, nx, ny, field1.vec_f, v);
              else
                varrayCopyMer(i, nx, ny, field1.vec_d, v);

              field2.vec_d[i] = percentile(v.data(), ny, p);
            }
          else
            {
              field2.vec_d[i] = missval;
              rnmiss++;
            }
        }
    }

  field2.nmiss = rnmiss;
}

void
field_mer_function(const Field &field1, Field &field2, int function)
{
  // clang-format off
  switch (function)
    {
    case func_min:    return field_mer_min(field1, field2);
    case func_max:    return field_mer_max(field1, field2);
    case func_range:  return field_mer_range(field1, field2);
    case func_sum:    return field_mer_sum(field1, field2);
    case func_meanw:  return field_mer_meanw(field1, field2);
    case func_avgw:   return field_mer_avgw(field1, field2);
    case func_stdw:   return field_mer_stdw(field1, field2);
    case func_std1w:  return field_mer_std1w(field1, field2);
    case func_varw:   return field_mer_varw(field1, field2);
    case func_var1w:  return field_mer_var1w(field1, field2);
    case func_kurt:   return field_mer_kurt(field1, field2);
    case func_skew:   return field_mer_skew(field1, field2);
    default: cdo_abort("%s: function %d not implemented!", __func__, function);
    }
  // clang-format on
}
